import json
import os
import subprocess

from . import config

ENV = os.environ.copy()
ENV['REGISTRY_AUTH_FILE'] = config.REGISTRY_AUTH_FILE

def run(cmd, **kwargs):
    return subprocess.run(cmd, check=True, env=ENV, **kwargs)

def out(cmd, **kwargs):
    return run(cmd, stdout=subprocess.PIPE, text=True, **kwargs).stdout.rstrip()

def get_apps():
    apps = {}
    data = json.loads(out(['podman', 'pod', 'ps', '--format', 'json']))
    for pod in data:
        app_name = pod['Labels'].get('spoc.app')
        app_version = pod['Labels'].get('spoc.version')
        if app_name:
            apps[app_name] = app_version
    return apps

def get_volumes_for_app(app_name):
    volume_ls = out(['podman', 'volume', 'ls', '--filter', f'label=spoc.app={app_name}',
                     '--format', 'json'])
    return set(volume['Name'] for volume in json.loads(volume_ls))

def start_pod(app_name):
    run(['podman', 'pod', 'start', app_name])

def stop_pod(app_name):
    run(['podman', 'pod', 'stop', '--ignore', app_name])

def get_pod_status(app_name=None):
    cmd = ['podman', 'pod', 'ps']
    if app_name:
        cmd.extend(['--filter', f'label=spoc.app={app_name}'])
    return out(cmd)

def create_volume(app_name, vol_name):
    run(['podman', 'volume', 'create', '--label', f'spoc.app={app_name}', vol_name])

def remove_volume(vol_name):
    run(['podman', 'volume', 'rm', vol_name])

def create_pod(app_name, app_version):
    run(['podman', 'pod', 'create', '--name', app_name,
         '--label', f'spoc.app={app_name}', '--label', f'spoc.version={app_version}'])

def remove_pod(app_name):
    stop_pod(app_name)
    run(['podman', 'pod', 'rm', '--ignore', app_name])

def create_container(app_name, cnt_name, image, env_file=None, volumes=None,
                     requires=None, hosts=None):
    cmd = ['podman', 'container', 'create', '--name', cnt_name, '--pod', app_name,
           '--restart', 'unless-stopped']
    if env_file:
        cmd.extend(['--env-file', env_file])
    if requires:
        cmd.extend(['--requires', ','.join(sorted(requires))])
    if volumes:
        for volume,mount in sorted(volumes.items(), key=lambda x: x[1]):
            cmd.extend(['--volume', f'{volume}:{mount}'])
    if hosts:
        for host in sorted(hosts):
            cmd.extend(['--add-host', f'{host}:127.0.0.1'])
    cmd.append(image)
    run(cmd)

def prune():
    run(['podman', 'image', 'prune', '--all', '--force'])
    run(['podman', 'volume', 'prune', '--force'])
